#!/usr/bin/perl

use strict ;
use warnings ;
use autodie ;

# # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### #
my $SERVER_NAME = $ENV{'SERVER_NAME'} ;
my $HELP_STRING = <<"EOS";
Returns all the non-expired PEM-encoded certificates from the certificate
database whose cn contains the passed parameter as a substring.

Examples:
   https://$SERVER_NAME/help
     returns the help text you are now reading

   https://$SERVER_NAME/certs/mdm
     returns certs for 'mdm.stanford.edu', 'mdm-dev1.stanford.edu',
     etc.

   https://$SERVER_NAME/certs/mdm.stanford.edu
     returns single certificate for 'mdm.stanford.edu'

   https://$SERVER_NAME/certs/mdm?newest=1
     returns whichever certificate expires latest among 'mdm.stanford.edu',
     'mdm-dev1.stanford.edu', etc.

   https://$SERVER_NAME/certs/mdm?unique=1
     returns 404 as there are more than one certificates containing 'mdm'

   https://$SERVER_NAME/certs/mdm^#?&*
     returns 400 as the search string should only contain characters valid
     for a CN

Options:
  newest=1 return the certificate with the latest expiration
  unique=1 fail if more than one certificate found
  help=1   show this help text

Notes:
Because the search is a substring match, more than one certificate might
be returned. To return only the certificate with the expiration date
farthest into the future, add the query parameter newest=1.

If you want the page to return an error if more than one certificate matches
cn, add the 'unique=1' query parameter. If BOTH newest=1 and unique=1
are present, unique takes precedence.

Errors:
If no certificate matches, returns HTTP_NOT_FOUND (404)
If 'unique' is set to 1 and multiple matches, returns HTTP_NOT_FOUND (404)
If CN passed in URI contains bad characters, HTTP_BAD_REQUEST (400)
EOS

# # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### #


use CGI ;
use HTTP::Status qw(:constants) ;

use Data::Dumper ;
use Stanford::Certificate ;

my $config_yaml_file = '/etc/certreq/cert-request' ;

my $q = CGI->new ;
# dump_env_and_exit() ;

# If the help option is set, return the help string.
if (wants_help($q)) {
    send_plain_text($q, $HELP_STRING) ;
} else {
    # Get the CN from the query parameter.
    my $cn = get_cn($q) ; # Might not return if cn contains bad characters.

    # Note that if there is an error, get_cert does not return.
    my $pems = get_cert($q, $cn) ;
    send_plain_text($q, $pems . "\n") ;
}

exit 0 ;

############################################################################################

sub get_cn {
    my ($q) = @_ ;

    if ($q->param('cn')) {
        my $cn = $q->param('cn') ;
        if ($cn !~ m{^[a-zA-Z0-9\-_\.]+$}xsm) {
            return_error(HTTP_BAD_REQUEST, "passed CN '$cn' is invalid") ;
        } else {
            return $cn ;
        }
    } else {
           return_error(HTTP_BAD_REQUEST, "missing parameter 'cn'") ;
    }
    return ; # Should never get here.
}

sub wants_newest {
    my ($q) = @_ ;
    return get_boolean_parameter($q, 'newest') ;
}

sub wants_unique {
    my ($q) = @_ ;
    return get_boolean_parameter($q, 'unique') ;
}

sub wants_help {
    my ($q) = @_ ;
    return get_boolean_parameter($q, 'help') ;
}

# Return 1 if the parameter is '1', 0 in all other cases.
sub get_boolean_parameter {
    my ($q, $param_name) = @_ ;

    if ($q->param($param_name)) {
        my $value = $q->param($param_name) ;
        if ($value eq '1') {
            return 1 ;
        }
    }
    return 0 ;
}

sub get_cert {
    my ($q, $cn) = @_ ;

    my $newest = wants_newest($q) ;
    my $unique = wants_unique($q) ;

    my ($schema, $cert) = make_connect() ;

    # Look for only those certificates that have not yet expired.
    my $now_MySQL = Stanford::Certificate::Misc::mysql_time() ;
    my $certs_rs = $schema->resultset('Certificate')->
        search(
            {
                -and =>
                [
                    'cn'         => {-like => "%$cn%" },
                    'expires_at' => { '>' => $now_MySQL },
                ]
            },
            {'order_by' => { -desc => 'expires_at' }}
        );

    # Keep count of how many certs come out.
    my $count = 0 ;
    my @pems = () ;
    while (my $cert = $certs_rs->next()) {
        push(@pems, $cert->pem()) ;
        ++$count ;
        if ($unique && ($count > 1)) {
            return_error(HTTP_NOT_FOUND,
                         qq{multiple certificates found matching } .
                         qq{'$cn' (unique parameter was set)}) ;
        }elsif ($newest && ($count == 1)) {
            return $pems[0] ;
        }
    }

    if ($count == 0) {
        return_error(HTTP_NOT_FOUND, "no certificate found matching '$cn'") ;
    } else {
        return join("\n", @pems) ;
    }

}

sub make_connect {
    # Set up the config global object.
    use Stanford::Certificate::Config ;
    my $CONFIG = Stanford::Certificate::Config::instance(
        'config_yaml_file' => $config_yaml_file
    ) ;

    my $schema = Stanford::Certificate::attach() ;
    my $cert = $schema->resultset('Certificate')->new({});

    return ($schema, $cert) ;
}

sub return_error {
    my ($http_error_code, $msg) = @_ ;
    print $q->header(-status=>$http_error_code,) ;
    print $msg . "\n" ;
    exit ;
}

sub dump_env_and_exit {
    my $q = CGI->new ;
    print $q->header() ;
    print '<pre>' . Dumper(%ENV) . '</pre>' ;
    exit 0 ;
}

sub send_plain_text {
    my ($q, $text) = @_ ;
    print $q->header('text/plain') ;
    print $text ;
    return ;
}
