package Stanford::Certificate::Config ;

use strict;
use warnings;

# Make this a Moose class
use Moose ;
use namespace::autoclean;

use Stanford::Orange::Util qw/ trim / ;
use Carp ;

my $ConfigSingleton ;

# The singleton pattern:
sub instance
{
  if (!$ConfigSingleton)
  {
    # warn "creating new config object\n" ;
    $ConfigSingleton = __PACKAGE__->new(@_) ;
  }

  return $ConfigSingleton ;
}

sub destroy
{
  $ConfigSingleton = undef ;
  return ;
}

# The master config file:
has config_yaml_file => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
  default   => '/etc/certreq/cert-request',
  trigger   => \&_config_yaml_file_validator,
) ;


sub _config_yaml_file_validator
{
  my ($self) = @_ ;

  if (!(-f $self->config_yaml_file()))
  {
    croak 'cannot find the config YAML file ' . $self->config_yaml_file() ;
  }
}



# The config attributes contained in config_yaml_file:
#
# use_prod_site:
#   Set to "YES" to use the production InCommon Certificate Mananger
#   site. Set "NO" (or any value not "YES") will cause all operations to
#   be made agains the test InCommon Certificate Mananger site.
#
# to_email_override:
#   If the to_email_override attribute is set to a non-emptry value, then
#   that value will be used as the "To:" address for _all_ outgoing
#   e-mails. This way, we can more easily test the system and avoid
#   sending annoying e-mails to real people. If it is set to an empty
#   value, it is ignored.
#
#   A useful override when testing is leland.test@mailtonews.stanford.edu
#   DEFAULT: undef
#
# from_email_address:
#  If this directive is set, use this address as the From: address for all
#  system-generated e-mail. If not set then will use the default
#  'nobody@stanford.edu'.
#
# incommon_cm_credentials_yaml_file :
#   The full path to the file containing the InCommon Certificate Manager
#   credentials.
#   DEFAULT: /etc/certreq/incommon_cm_credentials.yml
#
# db_connect_yaml_file :
#   The full path to the file containing the certificate request database
#   connection credentials.
#   DEFAULT: /etc/certreq/db_connect.yml
#
# incommon_approvers :
#   A comma-delimited list of approver e-mails. These e-mails should go to
#   people who can approve certificate requests at the InCommon Certificate
#   Manager web site.
#   DEFAULT: leland.test@mailtonews.stanford.edu
#
# validate_hostname_access_via_netdb:
#   We want to disable NetDB checking of whether the requester has
#   administrative access to the server in certain test cases. Set to
#   either "YES" or "NO"
#   DEFAULT: YES
#
# cc_warning_emails:
#   A comma-delimited list of e-mail addresses that are Cc'ed all
#   expiration warning e-mails.
#   DEFAULT: undefined


has use_prod_site => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has to_email_override => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
  trigger   => \&_set_to_email_override,
) ;
sub _set_to_email_override
{
  my ($self, $new_value, $old_value) = @_ ;

  # Trim the to_email_override
  # uncoverable branch false
  if (defined($self->to_email_override()))
  {
    $self->{'to_email_override'} = trim($new_value) ;
  }
  else
  {
    # uncoverable statement
    $self->{'to_email_override'} = q{} ;
  }

  return ;
}

has from_email_address => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => 'nobody@stanford.edu',
) ;

has incommon_cm_credentials_yaml_file => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => '/etc/certreq/incommon_cm_credentials.yml',
) ;

has incommon_approvers => (
  is        => 'rw',
  isa       => 'ArrayRef[Str]',
  required  => 1,
  default   => sub { ['leland.test@mailtonews.stanford.edu'] },
) ;

has db_connect_yaml_file => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => '/etc/certreq/db_connect.yml',
) ;

has validate_hostname_access_via_netdb => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 1,
) ;

has enable_validation => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 1,
) ;

has cc_warning_emails => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 0,
) ;


# Constructor: called AFTER object initial construction.
sub BUILD
{
  my $self = shift;
  my $args = shift ;

  use Data::Dumper ;

  $self->load_from_file() ;
}

sub to_s
{
  my $self = shift ;

  my $use_prod_site      = $self->use_prod_site() ;
  my $override           = $self->to_email_override() ;
  my $cm_credentials     = $self->incommon_cm_credentials_yaml_file() ;
  my $db_connect         = $self->db_connect_yaml_file() ;
  my $netdb              = $self->validate_hostname_access_via_netdb() ;
  my $config_file        = $self->config_yaml_file() ;
  my $enable_validation  = $self->enable_validation() ;
  my $from_email_address = $self->from_email_address() ;
  my $cc_warning_emails  = $self->cc_warning_emails() ;
  my @approvers          = @{ $self->incommon_approvers() } ;


  # ####  # ####  # ####  # ####  # ####  # ####  # ####  # ####
  my $format_bool_fref = sub
  {
    my ($value) = @_ ;
    if ($value == 1)
    {
      return 'YES' ;
    }
    # uncoverable branch false
    elsif ($value == 0)
    {
      return 'NO' ;
    }
    else
    {
      # uncoverable statement
      croak 'unrecognized boolean value' ;
    }
  } ;
  # ####  # ####  # ####  # ####  # ####  # ####  # ####  # ####

  # ####  # ####  # ####  # ####  # ####  # ####  # ####  # ####
  my $format_string_fref = sub
  {
    my ($value) = @_ ;
    if (!defined($value))
    {
      return '<NOT DEFINED>' ;
    }
    else
    {
      return $value ;
    }
  } ;
  # ####  # ####  # ####  # ####  # ####  # ####  # ####  # ####

  $enable_validation = $format_bool_fref->($enable_validation) ;
  $use_prod_site     = $format_bool_fref->($use_prod_site) ;

  my $incommon_approvers ;
  if (@approvers)
  {
    $incommon_approvers = join(q{, }, @approvers) ;
  }
  else
  {
    $incommon_approvers = 'NONE' ;
  }

  if (!$override)
  {
    $override = '<NONE>' ;
  }

  $cc_warning_emails = $format_string_fref->($cc_warning_emails) ;

  return <<"EOS";
config_yaml_file:                  $config_file
incommon_cm_credentials_yaml_file: $cm_credentials
use_prod_site:                     $use_prod_site
db_connect_yaml_file:              $db_connect
incommon_approvers:                $incommon_approvers
from_email_address:                $from_email_address
to_email_override:                 $override
enable_validation:                 $enable_validation
cc_warning_emails:                 $cc_warning_emails
EOS
}

sub load_from_file
{
  my $self = shift ;

  my $config_yaml_file = $self->config_yaml_file() ;

  # The configuration file should be in YAML format.
  use YAML::Tiny;

  # Create a YAML file
  my $yaml = YAML::Tiny->new;

  # Open the config
  $yaml = YAML::Tiny->read($config_yaml_file);
  if (!defined($yaml))
  {
    croak "failed to parse '$config_yaml_file' as a YAML file: "
        . YAML::Tiny->errstr() ;
  }

  $self->incommon_cm_credentials_yaml_file(
                       $yaml->[0]->{'incommon_cm_credentials_yaml_file'}) ;
  $self->db_connect_yaml_file(
                       $yaml->[0]->{'db_connect_yaml_file'}) ;

  my $use_prod_site =  $yaml->[0]->{'use_prod_site'} ;

  # If use prod site is not defined, default to FALSE.
  if (!defined($use_prod_site))
  {
    $self->use_prod_site(0) ;
  }
  if (defined($use_prod_site) && (trim($use_prod_site) =~ m{yes}i))
  {
    $self->use_prod_site(1) ;
  }
  else
  {
    $self->use_prod_site(0) ;
  }

  my $override = $yaml->[0]->{'to_email_override'} ;
  if (defined($override) && trim($override))
  {
    $self->to_email_override($override) ;
  }
  else
  {
    $self->to_email_override(q{}) ;
  }

  my $from_email_address = $yaml->[0]->{'from_email_address'} ;
  if (defined($from_email_address) && trim($from_email_address))
  {
    $self->from_email_address($from_email_address) ;
  }

  my $validate_hostname_access_via_netdb
    = $yaml->[0]->{'validate_hostname_access_via_netdb'} ;
  my $normalized ;

  # if validate_hostname_access_via_netdb is not defined, default to true
  # (1).
  if (!defined($validate_hostname_access_via_netdb))
  {
    $normalized = 1 ;
  }
  else
  {
    $validate_hostname_access_via_netdb =
        uc($validate_hostname_access_via_netdb) ;

    if ($validate_hostname_access_via_netdb eq 'YES')
    {
      $normalized = 1 ;
    }
    elsif ($validate_hostname_access_via_netdb eq 'NO')
    {
      $normalized = 0 ;
    }
    else
    {
      croak 'validate_hostname_access_via_netdb config parameter not understood: '
        . $validate_hostname_access_via_netdb ;
    }
  }
  $self->validate_hostname_access_via_netdb($normalized) ;

  my $approvers = $yaml->[0]->{'incommon_approvers'} ;
  my @approvers = split(q{,}, $approvers) ;
  my @incommon_approvers = () ;
  foreach my $approver (@approvers)
  {
    push (@incommon_approvers, trim($approver)) ;
  }

  $self->incommon_approvers(\@incommon_approvers) ;

  my $cc_warning_emails = $yaml->[0]->{'cc_warning_emails'} ;
  $self->cc_warning_emails($cc_warning_emails) ;

  return 1 ;
}

__PACKAGE__->meta->make_immutable;

1;
