package Stanford::Certificate::Misc ;

use strict;
use warnings;
use autodie qw(open close unlink) ;

require Exporter;
use base qw/ Exporter / ;

our @EXPORT_OK = qw(
  dedupe_array
  exit_with_error
  get_id
  log_message
  mysql_time
  mysql_time_local
  mysql_datetime_to_datecalc
);

use Capture::Tiny ;
use Date::Format ;
use Date::Parse;
use DateTime;
use Time::Zone;
use Date::Calc::Object;
use Math::BigInt ;
use Data::Dumper ;
use IPC::Run ;
use Carp;
use Memoize;

use Stanford::Orange::Util qw/
  trim
  tempfile
  slurp
  spew
  run_command
  run_command_improved
/ ;

use Stanford::PKObj ;
use Stanford::Certificate::Log ;

my %MONTH_NUMBER_TO_WORD =
(
  '1'  => 'Jan',
  '2'  => 'Feb',
  '3'  => 'Mar',
  '4'  => 'Apr',
  '5'  => 'May',
  '6'  => 'Jun',
  '7'  => 'Jul',
  '8'  => 'Aug',
  '9'  => 'Sep',
  '10' => 'Oct',
  '11' => 'Nov',
  '12' => 'Dec',
) ;



## Keep all time zones UTC!!!
## Keep all time zones UTC!!!
## Keep all time zones UTC!!!

sub mysql_time {
  my ($time_t) = @_ ;

  if (!$time_t)
  {
    # uncoverable statement
    $time_t = time() ;
  }
  my $template = '%Y-%m-%d %T' ;
  return time2str($template, $time_t, 'GMT');
}



# memoize tz_offset to shave a few milliseconds.
memoize('local_tz_offset');
sub local_tz_offset {
    return tz_offset('GMT') ;
}

# DateTime::TimeZone is SLOW, so make it once!
my $LOCAL_TZ ;
sub set_LOCAL_TZ {
    $LOCAL_TZ = DateTime::TimeZone->new( name => 'local' );
}

# Convert a MySQL GMT datetime to local time string, both in
# the YYYY-MM-DD HH:MM:SS format.
sub mysql_time_local {
  my ($mysql_gmt_time, $tz_string) = @_ ;

  # uncoverable branch true
  if (!defined(tz_offset('GMT')))
  {
    # uncoverable statement
    croak 'GMT is an unrecognized time zone string' ;
  }


  # If $LOCAL_TZ is NOT defined, but ENV{'TZ'} IS defined, set up $LOCAL_TZ
  # now.
  my $local_tz ;
  if ($tz_string) {
    $local_tz = DateTime::TimeZone->new( name => $tz_string );
  } elsif ($ENV{'TZ'} && (!$LOCAL_TZ)) {
      set_LOCAL_TZ() ;
  } else {
    $local_tz = DateTime::TimeZone->new( name => 'local' );
  }

  # Our MySQL times are all GMT. Ex: 2012-10-08 17:22:22
  if ($mysql_gmt_time =~ m{(\d\d\d\d)-(\d\d)-(\d\d)\ (\d\d):(\d\d):(\d\d)}xsm)
  {
    my $year   = $1 ;
    my $month  = $2 ;
    my $day    = $3 ;
    my $hour   = $4 ;
    my $minute = $5 ;
    my $secons = $6 ;

    my $dt = DateTime->new(
      year       => $1,
      month      => $2,
      day        => $3,
      hour       => $4,
      minute     => $5,
      second     => $6,
      time_zone  => 'GMT',
        );

    # Get the epoch (number of seconds since...)
    my $epoch = $dt->epoch() ;

    # Convert to a local time
    my $dt_local ;
    if ($local_tz) {
        $dt_local = DateTime->from_epoch('epoch' => $epoch, 'time_zone' => $local_tz) ;
    } elsif ($LOCAL_TZ) {
        $dt_local = DateTime->from_epoch('epoch' => $epoch, 'time_zone' => $LOCAL_TZ) ;
    } else {
        croak "cannot determine time-zone" ;
    }

    return $dt_local->strftime('%Y-%m-%d %T') ;
  }
  else
  {
    croak "could not parse MySQL time string '$mysql_gmt_time'" ;
  }
}

# Convert a MySQL datetime string to a date in the format '5-Oct-2012'
sub mysql_to_friendly_date
{
  my ($mysql_dt) = @_ ;

  if ($mysql_dt =~ m{^(\d\d\d\d)-(\d\d)-(\d\d)}xsm)
  {
    my $day   = q{} . ($3 + 0) ;
    my $month = q{} . ($2 + 0) ;
    my $year  = $1 ;

    # Convert month number to month letters
    my $month_word = $MONTH_NUMBER_TO_WORD{$month} ;
    return "${day}-${month_word}-${year}" ;
  }
  else
  {
    croak "could not parse MySQL date string '$mysql_dt'" ;
  }
}


# Convert a MySQL date-time string into a Date::Calc object.
sub mysql_datetime_to_datecalc
{
  my ($mysql_dt_string) = @_ ;

  # '2012-10-29 14:50:06'
  if ($mysql_dt_string =~ m{(\d\d\d\d)-(\d\d)-(\d\d)\ (\d\d):(\d\d):(\d\d)}xsm)
  {
    my $seconds = $6 ;
    my $minutes = $5 ;
    my $hours   = $4 ;
    my $dom     = $3 ;
    my $month   = $2 ;
    my $year    = $1 ;

    return Date::Calc->new($year, $month, $dom, $hours, $minutes, $seconds) ;
  }
  else
  {
    croak "could not parse MySQL date string '$mysql_dt_string'" ;
  }
}

# Returns the result of the incommon-mgr status call. Currently, that
# is one of:
# * 'ready' (certificate ready for downloading)
# * 'not ready: approved but not yet fulfilled'
# * 'not ready: submitted but not yet approved'
# * 'not ready: declined'
sub status
{
  my ($order_number, $credentials_file, $test) = @_ ;

  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####
  my $check_nonempty_fref = sub
  {
    my ($value, $label) = @_ ;
    if (!$value)
    {
      croak "cannot collect without a $label" ;
    }
    return ;
  } ;
  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####
  $check_nonempty_fref->($order_number,     'order number') ;
  $check_nonempty_fref->($credentials_file, 'credentials file') ;

  if (!(-e $credentials_file))
  {
    croak "credentials file '$credentials_file' does not exists or is not readable" ;
  }

  # Get the status of the request using incommon-mgr
  my @cmd = ('incommon-mgr',
             'status',
             $order_number,
             '--credentials-file', $credentials_file,
      ) ;

  if ($test) {
    push(@cmd, '--test') ;
  }

  my ($stdout, $stderr, $exit_code) = run_command_improved(@cmd) ;

  if ($stderr) {
      my $cmd_string = join(q{ }, @cmd) ;
      croak "command '$cmd_string' returned an error: $stderr" ;
  } else {
      my $rv = trim($stdout) ;

      return $rv ;
  }
}


# Returns a hash.
# The hash has two keys:
#   'renewid' --> the renew id string.
#   'pems'    --> a reference to an array containing the PEM strings.
sub collect
{
  my ($order_number, $credentials_file, $test) = @_ ;

  my %result_hash = () ;
  $result_hash{'renewid'} = undef ;

  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####
  my $check_nonempty_fref = sub
  {
    my ($value, $label) = @_ ;
    if (!$value)
    {
      croak "cannot collect without a $label" ;
    }
    return ;
  } ;
  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####  # ## ####
  $check_nonempty_fref->($order_number,     'order number') ;
  $check_nonempty_fref->($credentials_file, 'credentials file') ;

  if (!(-e $credentials_file))
  {
    croak "credentials file '$credentials_file' does not exists or is not readable" ;
  }

  # Get the cert using incommon-mgr

  my @cmd = ('incommon-mgr',
             'collect',
             $order_number,
             '--credentials-file', $credentials_file,
      ) ;

  if ($test) {
      push(@cmd, '--test') ;
  }

  my ($stdout, $tderr, $exit_code) = run_command_improved(@cmd) ;
  my $rv = trim($stdout) ;

  if (!$rv) {
    croak 'collect did not return certificate bundle' ;
  }
  else
  {
    # The string returned should end with "renewid: XXXX" where "XXXX" is
    # the renew ID. We want to grab that.
    # renewID: IxNXKItVO4t6UAE9NtjP
    if ($rv =~ m{renewID:\ (\S+)\s*}xsm)
    {
      $result_hash{'renewid'} = $1 ;

      # Remove the renewid line from the string.
      $rv =~ s{renewID:\ (\S+)\s*}{}xsm ;
      $rv = trim($rv) ;
    }

    my @pem_strings = Stanford::Certificate::X509::split_pems_string($rv) ;
    $result_hash{'pems'} = \@pem_strings ;

    return %result_hash ;
  }
}

sub log_message {
  my (%args) = (
                'request_id'     => undef,
                'certificate_id' => undef,
                @_,
               ) ;

  my $schema = $args{'schema'} ;
  if (!$schema)
  {
    croak 'cannot log a message without a schema' ;
  }

  # Make new eventlog object
  my $eventlog = $schema->resultset('Eventlog')->new({});

  my $level   = $args{'level'} ;
  my $message = $args{'message'} ;

  if (defined($level))
  {
    $eventlog->level($level) ;
  }

  $eventlog->message($message) ;
  $eventlog->insert() ;

  return $eventlog ;

}

sub dump_all_eventlogs
{
  my ($schema) = @_ ;
  my $eventlogs_rs = $schema->resultset('Eventlog') ;
  return Stanford::Certificate::Result::Eventlog::format_eventlog_rs($eventlogs_rs) ;
}


# Get the InCommon/Comodo Certificate Manager website username and
# password. If no credentials file is passed, use the one stored in the
# the config object.
sub get_username_password
{
  my ($credentials_file) = @_ ;

  # If no credentials file passed, use a reasonable default.
  if (!$credentials_file)
  {
    my $config = Stanford::Certificate::Config::instance() ;
    $credentials_file = $config->incommon_cm_credentials_yaml_file() ;
  }

  if (!(-e $credentials_file))
  {
    croak "could not find credentials file '$credentials_file'" ;
  }

  use YAML::Tiny;
  my $yaml = YAML::Tiny->new;
  $yaml = YAML::Tiny->read($credentials_file);

  my $username = $yaml->[0]->{username} ;
  my $password = $yaml->[0]->{password} ;

  return ($username, $password) ;
}



# This is a utility function only used by the utility scripts (popluate.pl
# and stanford_cert_make_schema.pl)

# uncoverable subroutine
sub get_connect_yml_file
{
  my $platform = $ARGV[0] ;
  # If no platform is passed assume 'production'
  if (!$platform)
  {
    $platform = 'production' ;
  }
  elsif ($platform =~ m{test}ism)
  {
    $platform = 'testing' ;
  }

  use FindBin;
  FindBin::again();

  my %platform_to_file =
  (
    # This path should be relative to the utility scripts directory bin/
    'testing' => $FindBin::Bin . '/../sslcert_db_testing.yml',
    'uat'     => $FindBin::Bin . '/../sslcert_db_uat.yml',
    'sbx'     => $FindBin::Bin . '/../sslcert_db_sbx.yml',
    'production' => $FindBin::Bin . '/../sslcert_db_prod.yml',
  ) ;

  my $connect_yml_file_default = '/etc/certreq/cert-request-db.yml' ;
  my $connect_yml_file = $connect_yml_file_default ;
  if (exists($platform_to_file{$platform}))
  {
    $connect_yml_file = $platform_to_file{$platform} ;
  }

  return $connect_yml_file ;
}



# Mark any approved or fulfilled requests.
sub update_all {
    my ($schema, $incommon_cm_credentials_file, $test) = @_ ;

    if (!$incommon_cm_credentials_file) {
        croak 'cannot update without a credentials file' ;
    }

    if (!defined($test)) {
        croak 'the test parameter must be defined' ;
    }

    my $rv = q{} ;

    my $requests_rs ;

    # Get all the certificates that have an order number but whose "approved"
    # field is FALSE and that are NOT cancelled.
    $requests_rs = $schema->resultset('Request')
        ->search_literal('(order_number IS NOT NULL) ' .
                         'AND (approved = 0) AND (cancelled = 0)')
          ;

    ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###
    my $logger_fun = sub {
        my ($msg) = @_ ;
        return Stanford::Certificate::Log::INFO('[update_all] ' . $msg) ;
    } ;
    ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###
    # Get the status but wrap in an eval in case the get status API call
    # fails. If it does fail, log this as an error.
    my $get_status_fun = sub {
        my ($request) = @_ ;

        my $tmp_status ;
        eval {
            $tmp_status = $request->status($incommon_cm_credentials_file) ;
        } ;
        my $status_at_error = $@ ;

        if ($status_at_error) {
            my $id = $request->id() ;
            my $cn = $request->cn() ;
            my $msg = "error getting status for certificate $id ($cn)" ;
            Stanford::Certificate::Log::ERROR('[update_all] ' . $msg) ;
        }
        return ($tmp_status, $status_at_error) ;
    } ;
    ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###


    my $number_requests_pending_approval = 0 ;
    while (my $request = $requests_rs->next()) {
        $request->test($test) ;
        ++$number_requests_pending_approval ;
        my $order_number = $request->order_number() ;
        my $cn           = $request->cn() ;
        my $id           = $request->id() ;

        # Get the status but if the get status results in an error, skip.
        my ($status, $err_message) = $get_status_fun->($request) ;
        if ($err_message) {
            next ;
        }

        # Is this approved yet?
        if ($status =~ m{approved|fulfilled}xsm) {
            $request->approved(1) ;
            $request->update() ;
            my $msg = "Request for '$cn' has been marked approved (db id: $id)" ;
            $rv .= "$msg\n" ;
            $logger_fun->($msg) ;
        } elsif ($status =~ m{declined}xsm) {
            # This request has been declined by an admin.
            $request->declined(1) ;
            $request->cancelled(1) ;
            $request->update() ;
            my $msg = "Request for '$cn' has been declined (db id: $id)" ;
            $logger_fun->($msg) ;
        } else {
            # This certificate not yet approved.
            my $msg = "Request for '$cn' still awaiting approval (db id: $id)" ;
            $logger_fun->($msg) ;
        }
    }

    if ($number_requests_pending_approval == 0) {
        my $msg = 'No certificate requests awaiting approval' ;
        $logger_fun->($msg) ;
    }

    # Get all the certificates that have been approved but do not yet have a
    # certificate id and are NOT cancelled.
    $requests_rs = $schema->resultset('Request')
        ->search_literal('(approved = 1) AND (cancelled = 0) AND (declined = 0) AND (certificate_id IS NULL)')
        ;

    my $number_requests_pending_fulfillment = 0 ;
    while (my $request = $requests_rs->next()) {
        $request->test($test) ;
        ++$number_requests_pending_fulfillment ;
        my $order_number = $request->order_number() ;
        my $cn           = $request->cn() ;
        my $id           = $request->id() ;

        # Get the status but if the get status results in an error, skip.
        my ($status, $err_message) = $get_status_fun->($request) ;
        if ($err_message) {
            next ;
        }

        if ($status =~ m{fulfilled}sm) {
            # Collect
            my $certificate_id = $request->collect($incommon_cm_credentials_file) ;
            my $msg = "Request for $cn downloaded (db id: $id)" ;
            $logger_fun->($msg) ;
            $rv .= "$msg\n" ;
        } elsif ($status =~ m{declined}xsm) {
            # This request has been declined by an admin. (This should
            # only happen in the rare case when a request has been
            # approved but then declined before the system gets a chance
            # to issue it.)
            $request->declined(1) ;
            $request->cancelled(1) ;
            $request->update() ;
            my $msg = "Request for '$cn' has been declined (db id: $id)" ;
            $logger_fun->($msg) ;
        } else {
            my $msg = "Request for $cn approved but still not fulfilled (db id: $id ($status))" ;
            $logger_fun->($msg) ;
        }
    }

    if ($number_requests_pending_fulfillment == 0) {
        my $msg = 'No certificate requests awaiting fulfillment' ;
        $logger_fun->($msg) ;
    }

    return $rv ;
}

# Query the InCommon Certificate Manager for a list of all valid
# certificate types. Return an array of certificate type names.
sub cert_types
{
  my ($incommon_cm_credentials_file, $test) = @_ ;

  if (!defined($test))
  {
    croak 'the test parameter MUST be provided' ;
  }

  if (!$incommon_cm_credentials_file)
  {
    croak 'missing credentials file' ;
  }


  my @cmd = ('incommon-mgr',
             'cert-types',
             '--credentials-file', $incommon_cm_credentials_file,
      ) ;

  if ($test) {
    push(@cmd, '--test') ;
  }

  my ($stdout, $stderr, $exit_code) = run_command_improved(@cmd) ;
  my $rv = $stdout ;

  my @lines = split(/\n/xsm, $rv) ;
  return @lines ;
}


# One digit pad
sub pad1
{
  my ($n) = @_;
  if ($n + 0 < 10) { return '0' . $n ; }
  else             { return q{} . $n ; }
}

# Remove duplicates from an array of strings.
# Returns the de-duped array.
sub dedupe_array
{
  my ($array_ref) = @_ ;

  my %found = () ;
  my @elements_nodupes = () ;
  foreach my $element (@{$array_ref})
  {
    if ($found{$element})
    {
      next ;
    }
    else
    {
        push(@elements_nodupes, $element) ;
        $found{$element} = 1 ;
    }
  }

  return @elements_nodupes ;
}


# Returns the tuple ($id, $found_object) where $id is the id of the
# certificate found, and $found_object is the Certificate object of the
# found object.
sub get_id {
    my ($schema, $type, $id, $command_name) = @_ ;

    if (!$command_name) {
        $command_name = 'UNKNOWN' ;
    }

    # If no id, exit with an error.
    if (!$id) {
        exit_with_error("the $command_name command requires a database id") ;
    } elsif ($id !~ m{^\d+L?$}xsm) {
        exit_with_error("the $command_name database id should be a number or a number followed by 'L'") ;
    }

    # Is this a legacy expiration id?
    my $legacy_expiration ;
    if ($id =~ m{L}ixsm) {
        $legacy_expiration = 1 ;
    } else {
        $legacy_expiration = 0 ;
    }

    # Get the numerical portion
    $id =~ s{L}{}gxsm ;

    my $found_object ;
    if ($legacy_expiration) {
        # Get the legacy expiration object
        my $leg = $schema->resultset('LegacyExpiration')->new({}) ;
        $found_object = $leg->findById($id) ;

        if (!$found_object) {
            exit_with_error("could not find legacy expiration object with database id '$id'") ;
        }
    } elsif ($type =~ m{req}ixsm) {
        # Get the request.
        my $req       = $schema->resultset('Request')->new({}) ;
        $found_object = $req->findById($id) ;
        if (!defined($found_object)) {
            exit_with_error("could not find request with database id '$id'") ;
        }
    } elsif ($type =~ m{cert}ixsm) {
        # Get the request.
        my $cert      = $schema->resultset('Certificate')->new({}) ;
        $found_object = $cert->find_by_id($id) ;
        if (!defined($found_object)) {
            exit_with_error("could not find certificate with database id '$id'") ;
        }
    } else {
        croak "unknown type '$type'" ;
    }

    return ($id, $found_object) ;
}

sub get_request_by_order_number {
    my ($schema, $order_number) = @_ ;

    # If no order number, exit with an error.
    if (!$order_number) {
        exit_with_error('missing required order number parameter') ;
    }

    # Get the request.
    my $req       = $schema->resultset('Request')->new({}) ;
    my $found_object = $req->findByOrderNumber($order_number) ;

    if (!defined($found_object)) {
        exit_with_error(qq{could not find request with order number '$order_number'}) ;
    } else {
        my $id = $found_object->id() ;
        return ($id, $found_object) ;
    }
    return ; # Should never get here.
}

sub exit_with_error {
    my ($msg) = @_ ;
    print "error: $msg\n" ;
    exit 1 ;
}

# Load InCommon statuses from a JSON file.
sub load_incommon_statuses {
    my ($schema, $filename, $progress) = @_ ;

    my $quiet ;

    ## # ####    ## # ####    ## # ####    ## # ####    ## # ####
    my $progress_null_fref = sub {
        return ;
    } ;
    my $progress_stdout_fref = sub {
        my ($msg) = @_ ;
        print {*STDOUT} $msg . "\n" ;
        return ;
    } ;
    my $progress_stderr_fref = sub {
        my ($msg) = @_ ;
        print {*STDERR} $msg . "\n" ;
        return ;
    } ;
    ## # ####    ## # ####    ## # ####    ## # ####    ## # ####

    my $progress_fref  ;
    if ($progress =~ m{quiet}ism) {
        $quiet = 0 ;
        $progress_fref = $progress_null_fref ;
    } else {
        $quiet = 1 ;
        if ($progress =~ m{stdout}sm) {
            $progress_fref = $progress_stdout_fref ;
        } elsif ($progress =~ m{stderr}sm) {
            $progress_fref = $progress_stderr_fref ;
        } else {
            croak "unknown progress parameter '$progress'" ;
        }
    }


    import JSON ;
    my $x = slurp($filename) ;

    my $json = JSON->new->allow_nonref();

    my $statuses_aref = $json->decode( $x ) ;

    my $cert = $schema->resultset('Certificate')->new({});

    # ## ###    # ## ###    # ## ###    # ## ###    # ## ###    # ## ###    # ## ###
    my $add_status_fref = sub {
        my ($st_href, $certif, $field, $status) = @_  ;

        if (exists($st_href->{$field})) {
            my $status_at = $st_href->{$field} ;
            $status_at =~ s{(\d\d:\d\d:\d\d)\.\d\d\d\d\d\d}{$1}xsm ;
            my $inc_stat = $schema->resultset('IncommonStatus')->new({});
            $inc_stat->status($status) ;
            $inc_stat->status_at($status_at) ;
            if ($certif->add_status($inc_stat)) {
                $progress_fref->(sprintf('added status %9s for certificate %s',
                                        $status, $certif->cn())) ;
                $certif->update() ;
            } else {
                # Nothing to do since it already has the status.
            }
        }
        return ;
    } ;
    # ## ###    # ## ###    # ## ###    # ## ###    # ## ###    # ## ###    # ## ###

    use Data::Dumper ;
    foreach my $status_href (@{$statuses_aref}) {
        my $status       = lc($status_href->{'status'}) ;
        my $order_number = $status_href->{'id'} ;

        # If we have a declined or rejected status, we need to update the request object.
        if ($status =~ m{declined|rejected}xsm) {
            # Find the request. Trap in case there is a dupe.
            my $request = $schema->resultset('Request')->new({}) ;
            my $request1 ;
            eval {
                $request1 = $request->findByOrderNumber($order_number) ;
            } ;
            my $eval_error = $@ ;
            if ($eval_error) {
                if ($eval_error =~ m{found\ more\ than\ one\ request}xsm) {
                    $progress_fref->($eval_error) ;
                } else {
                    croak $eval_error ;
                }
            }

            if (!$request1) {
                # $progress_fref->("no request found for order number '$order_number'") ;
            } else {
                my $cn = $request1->cn() ;
                if ($status =~ m{declined}xsm) {
                    if (!$request1->declined()) {
                        $progress_fref->("marking request $order_number ($cn) as 'declined'") ;
                        $request1->declined(1) ;
                    }
                } elsif ($status =~ m{rejected}xsm) {
                    if (!$request1->rejected()) {
                        $progress_fref->("marking request $order_number ($cn) as 'rejected'") ;
                        $request1->rejected(1) ;
                    }
                } else {
                    croak 'do not know what to do here' ;
                }

                $request1->update() ;
            }
        }

        # Find the certificate (if there is one). Trap in case there is a dupe
        my $cert1 ;
        eval {
            $cert1 = $cert->find_by_order_number($order_number) ;
        } ;
        my $eval_error = $@ ;
        if ($eval_error) {
            if ($eval_error =~ m{found\ more\ than\ one\ request}xsm) {
                if (!$quiet) {
                    carp $eval_error ;
                }
            } else {
                croak $eval_error ;
            }
        }

        if (!$cert1) {
            # warn "could not find certificate with request $order_number" ;
        } else {
            # We have the certificate. Add the incommon status.

            $add_status_fref->($status_href, $cert1, 'approved',  'approved') ;
            $add_status_fref->($status_href, $cert1, 'revoked',   'revoked') ;
            $add_status_fref->($status_href, $cert1, 'requested', 'requested') ;
            $add_status_fref->($status_href, $cert1, 'issued',    'issued') ;

            # warn $cert1->to_s() ;
        }
    }

    return ;
}

# Return an array of Stanford::PKObj objects.
sub download_live_certificates {
    my ($cn) = @_ ;

    use Net::SSLeay qw/ sslcat / ;

    # Download cert, but set a time-out.
    my $timeout = 10 ;
    my ($reply, $err, $cert) ;
    eval {
        local $SIG{ALRM} = sub { die "alarm\n" }; # NB: \n required
        alarm($timeout);
        ($reply, $err, $cert) = sslcat($cn, 443, q{/}) ; # 5
        alarm(0);
    };
    my $at_error = $@ ;
    if ($at_error && ($at_error ne "alarm\n")) {
        croak $at_error ; # propagate unexpected errors
    } else {
        $err = 'certificate download timed-out' ;
    }

    if (!$cert) {
        croak "could not download certificate for '$cn': $err" ;
    }

    my $PEM = Net::SSLeay::PEM_get_string_X509($cert) ;

    if (!$PEM) {
        croak "no certificate extracted for '$cn'" ;
    }

    # Parse the output for PKI objects.
    my $fh1        = File::Temp->new(UNLINK => 0) ;
    my $pem_file   = $fh1->filename() ;
    spew($PEM, $pem_file) ;

    my $pkobj = Stanford::PKObj->new('file' => $pem_file) ;
    my @pkobjects = $pkobj->load_from_file() ;

    if (-e $pem_file) {
        unlink($pem_file) ;
    }

    return @pkobjects ;
}

# Get the certificate. If none match, return undef.
#
# If providing $subject_regex, this function will escape any
# metacharacters, so do NOT do any escaping yourself.
sub download_live_certificate {
    my ($cn, $subject_regex) = @_ ;

    if (!$subject_regex) {
        $subject_regex = qr/CN=$cn/xsm ;
    } else {
        # Use \Q..\E to escape all the regex metacharacters
        $subject_regex = qr/\Q$subject_regex\E/xsm ;
    }

    my @pkobjs = download_live_certificates($cn) ;
    foreach my $pkobj (@pkobjs) {
        my %field_to_value = Stanford::PKObj::parse_x509($pkobj->get_contents()) ;
        my $subject = $field_to_value{'subject'} ;
        if ($subject =~ $subject_regex) {
            return $pkobj ;
        }
    }

    return undef ;
}



1;
