package Stanford::Certificate::Notification ;

use strict;
use warnings;

use Stanford::Certificate::Result::Request ;
use Stanford::Certificate::Misc qw(mysql_datetime_to_datecalc) ;
use Stanford::Certificate::Mail ;
use Carp ;

# Return a list of request objects for certificates that are expiring in
# at least M but not more than N days from the date passed in as the first
# argument. This list is sorted first by the cn (common name), and then by
# the expiration date.
#
# Arguments:
#
# $request: any connected Stanford::Certificate::Result::Request object.
#
# $dc: A Date::Calc::Object object set to the GMT date and time you are
# calculating with respect to (normally it is set to the current time).
#
# $lower: The lower limit in days that you want to filter on.
#
# $upper: The upper limit in days that you want to filter on.
#
# Example:
#
#    my $dc_now = Date::Calc->gmtime() ;
#    expiring_requests($request, $dc_now, 2, 3)
#
# will return all the request objects whose associated certificates are
# expiring in more than 48 hours but less than 72 hours.
#
# If the third argument is omitted, then $upper defaults to $lower + 1.
#
# Although unlikely, for a given subject this function _could_ return more
# than one request for that subject. For example, if two certificates with
# close expiration dates and the same subject were provisioned.
#
# WARNING! The expires_at and valid_at values from the database are stored
# in GMT, so be sure that the $dc object is also in GMT.
sub expiring_requests
{
  my ($schema, $dc, $lower, $upper) = @_ ;

  if (!defined($upper))
  {
    $upper = $lower + 1 ;
  }

  # Make copies of the passed in Date::Calc object.
  my $date_lower = $dc->clone();
  my $date_upper = $dc->clone();

  # Add the lower and upper number of days.
  $date_lower += [0, 0, $lower, 0, 0, 0] ;
  $date_upper += [0, 0, $upper, 0, 0, 0] ;

  # Make MySQL string versions of these dates.
  my $date_MySQL_lower = $date_lower->string(\&datetime_formatter_mysql) ;
  my $date_MySQL_upper = $date_upper->string(\&datetime_formatter_mysql) ;
  # warn "\$date_MySQL_lower: $date_MySQL_lower; \$date_MySQL_upper: $date_MySQL_upper";

  # Get all requests that have a non-null certificate_id and whose
  # associated certificate expires in the range ($date_lower, $date_upper).
  # Note that this might return multiple requests for the same certificate
  # (e.g., renewals), so we order first by subject and then by increasing
  # expires_at (so the certificates expiring closest to now come first).

  #$schema->storage->debug(1);
  # Note that in a table join, the search object table (in this case
  # 'Request') always gets aliased to 'me'.
  my @requests = $schema->resultset('Request')->search(
           {
             -and =>
             [
               'me.certificate_id'         => { '!=', undef },
               'me.suppress_notifications' => { '=', 0 },
               'certificate.expires_at' => { '<' => $date_MySQL_upper },
               'certificate.expires_at' => { '>' => $date_MySQL_lower },
             ]
           },
           {
             join     => 'certificate',
             order_by =>
               [
                 { -asc  => [qw/certificate.cn/] },
                 { -desc => [qw/certificate.expires_at/] },
               ],
           })->all() ;

  #$schema->storage->debug(0);
  #             -desc => [qw/certificate.cn certificate.expires_at/]

  return @requests ;
}

# This is a special function that takes a list of requests and, if there
# are succesive requests with certificates with the same CN, it keeps only
# the first request.
#
# IMPORTANT: We assume that the requests are presented in the sort order
# that is returned by the expiring_requests function.
#
# The use case is that there is a certificate that is expiring soon but
# someone has added a renewal of that certificate in the database. In this
# case, we don't want to send a message about the certificate that is
# expiring soon since there is already a renewed certificate.
#
#   my @filtered_requests = unique_latest_requests(\@requests) ;
#   # @filtered_requests will be a sub-array of @requests with no
#   # duplications of cn.
#
sub unique_latest_requests
{
  my ($requests_aref) = @_ ;
  my @requests = @$requests_aref ;

  my $last_cn = undef ;
  my @results = () ;
  foreach my $request (@requests)
  {
    # Get the certificate associated with this request.
    my $certificate = $request->certificate() ;
    my $current_cn = $request->cn() ;
    # warn "($last_cn, $current_cn)" ;

    # Is the current $cn the same as the last one? If so, we do NOT add it
    # to the @results arrays. If the
    if (!defined($last_cn))
    {
      push(@results, $request) ;
      $last_cn = $current_cn ;
    }
    elsif ($last_cn eq $current_cn)
    {
      # We _don't_ add the current request to the list as it has the same
      # CN as the last request and hence an _earlier_ expiration date.
    }
    else
    {
      # The current CN is for a new certificate, so we add it to the
      # @results array and update $last_cn.
      push(@results, $request) ;
      $last_cn = $current_cn ;
    }
  }

  return @results ;
}

# Given an array of requests, remove any that have a CN that is associated
# with another request and that request's certificate has an expiration
# date _later_ than the first request.
#
# Algorithm: let $r0 be a member of the passed-in array @requests and
# let $c0 be its associated certificate.
#
#  1. Let $CN be the common name for $c0.
#  2. Let $c1, $c2, ..., $cn be all certificates in the database whose common
#     name is _also_ $CN.
#  3. If at least one of $c1, $c2, ..., $cn has an expiration date
#     strictly LATER than $c0, then remove $r0 from the results
#     array. Otherwise, add it to the results array.
sub requests_with_no_later
{
  my ($requests_aref) = @_ ;
  my @requests = @$requests_aref ;

  my @filtered_results = () ;
  foreach my $request (@requests)
  {
    # 1. Get the common name for $request.
    my $cn = $request->cn() ;
    if (!$cn)
    {
      die "this request has no common name!" ;
    }

    # 2. Get all certificates with the common name $cn.
    my $certificate = $request->certificate() ;
    my @matching_certificates = $certificate->find_by_cn($cn) ;

    my $found_later = 0 ; # tells us if we found a later expiring cert.
    if (scalar(@matching_certificates) == 0)
    {
      # We know there is at least one, so if we get zero this is a bug.
      die "found no certificates with cn '$cn'; should have been at least one" ;
    }
    else
    {
      my $dc_certificate_expires_at
          = mysql_datetime_to_datecalc($certificate->expires_at()) ;

      # See if any have a date later than $certificate.
      CERTS: foreach my $matching_certificate (@matching_certificates)
      {
        my $dc_matching_expires_at
          = mysql_datetime_to_datecalc($matching_certificate->expires_at()) ;
        if ($dc_matching_expires_at > $dc_certificate_expires_at)
        {
          # Found one that has a later expiration date, so can stop looking.
          $found_later = 1 ;
          last CERTS ;
        }
      }
    }

    # If we did NOT find a certificate with a later expiration date, add
    # $certificate to the results array.
    if (!$found_later)
    {
      push (@filtered_results, $request) ;
    }
  }

  return @filtered_results ;
}

# Send e-mails to all contacts for requests whose certificates are
# expiring in N days.
#
# Example:
#   send_expiration_warning_emails(7)
#
# will send an expiration warning e-mail to all contacts of requests
# that have certificates that are expiring between 7 * 24 and 8 * 48 hours
# from the current time.
#
# Note that this function does NOT send warning e-mails to legacy
# certificates.
#
# Returns the number of requests sent e-mail notifications.
sub send_expiration_warning_emails
{
  my ($request, $number_of_days) = @_ ;

  if ($number_of_days < 1)
  {
    die "the number of days must be an integer at least as large as 1" ;
  }

  # We need a Date::Calc object for the current Zulu time.
  my $dc_now = Date::Calc->gmtime() ;

  # 1. Get all requests that have certificates expiring in between N and
  #    N+1 days.
  my @requests = $request->expiring_requests($dc_now, $number_of_days, $number_of_days + 1) ;

  # 2. Remove any duplicate requests. That is, leave only one request for each
  # CN, that request with the later (or latest) expiration date.
  @requests = Stanford::Certificate::Notification::unique_latest_requests(\@requests) ;

  # 3. Finally, remove any requests that have a CN that is assocated with
  # a second request and that second request's certificate is NOT expiring.
  @requests = Stanford::Certificate::Notification::requests_with_no_later(\@requests) ;

  # 4. We are now ready to send the warning e-mails.
  foreach my $request (@requests)
  {
    $request->send_expiration_warning_email() ;
  }

  return scalar(@requests) ;
}

# This is similar to send_expiration_warning_emails except we only send
# e-mails to legacy certificates, that is, certificates listed in the
# legacy_expirations table.
#
# Because LegacyExpirations use dates rather than datetimes, we return all
# LegacyExpirations that have an expiration greater than or equal to $dc +
# $lower but strictly less than $dc + $upper.
#
# Returns a list of the legacy expirations that should be notified.
sub get_expiring_legacy_expirations
{
  my ($schema, $dc, $lower, $upper) = @_ ;

  if (!$dc)
  {
    $dc = Date::Calc->gmtime() ;
  }

  if (!defined($upper))
  {
    $upper = $lower + 1 ;
  }

  # Make copies of the passed in Date::Calc object.
  my $date_lower = $dc->clone();
  my $date_upper = $dc->clone();

  # Add the lower and upper number of days.
  $date_lower += [0, 0, $lower, 0, 0, 0] ;
  $date_upper += [0, 0, $upper, 0, 0, 0] ;

  # Make MySQL string versions of these dates.
  my $date_MySQL_lower = $date_lower->string(\&date_formatter_mysql) ;
  my $date_MySQL_upper = $date_upper->string(\&date_formatter_mysql) ;

  my @l_expirations = $schema->resultset('LegacyExpiration')->search(
           {
             -and =>
             [
               'suppress_notifications' => { '=', 0 },
               'expires_at' => { '<'  => $date_MySQL_upper },
               'expires_at' => { '>=' => $date_MySQL_lower },
             ]
           },
          )->all() ;

  foreach my $l_expiration (@l_expirations)
  {
    my $cn = $l_expiration->cn() ;
  }

  return @l_expirations ;
}

# Given an array of legacy expiration objects, mails each of them
# an expiration notice.
sub email_expiring_legacy_expirations
{
  my ($legacy_expirations_aref) = @_ ;

  my @legacy_expirations = @{$legacy_expirations_aref} ;
  foreach my $legacy_expiration (@legacy_expirations)
  {
    $legacy_expiration->send_expiration_warning_email() ;
  }
}


# Convert a Date::Calc object into a MySQL datetime string.
sub datecalc_to_mysql
{
  my ($dc) = @_ ;
  return $dc->string(\&datetime_formatter_mysql) ;
}

sub pad
{
  my ($n) = @_;
  if ($n + 0 < 10) { return '0' . $n ; }
  else             { return q{} . $n ; }
}

sub datetime_formatter_mysql
{
  my($self,$code,$lang) = @_;

  my $year  =     $self->year() ;
  my $month = pad($self->month()) ;
  my $day   = pad($self->day()) ;
  my $hour  = pad($self->hours()) ;
  my $min   = pad($self->minutes()) ;
  my $sec   = pad($self->seconds()) ;

  return qq{$year-$month-$day $hour:$min:$sec} ;
} ;

sub date_formatter_mysql
{
  my($self,$code,$lang) = @_;

  my $year  =     $self->year() ;
  my $month = pad($self->month()) ;
  my $day   = pad($self->day()) ;

  return qq{$year-$month-$day} ;
} ;




# Given a Date::Calc object, returns another Date::Calc object date N days
# from the given date.
#
# Example: if $dc is Nov 29, 2012 and N is 1, then this function returns
# a Date::Calc object for Nov 30, 2012. If N is 2, then it returns
# a Date::Calc object for Dec 1,  2012, etc.
#
#
# $dc is a Date::Calc object.
# $N is the number of days. It will be truncated to the nearest integer;
# if it is not at least 1 then raise an exception.
sub days_from_now
{
  my ($dc, $N, $at_least_24_hours) = @_ ;

  # Get the date $number_of_days from now.
  my $dc_future = $dc + [0, 0, $N, 0, 0, 0] ;

  return $dc_future ;
}


sub send_warning_email
{
  my ($cn_uppercase, $expires_at_date_local, $contact_email) = @_ ;

  my $config = Stanford::Certificate::Config::instance() ;
  my $from_email_address = $config->from_email_address() ;

  # If the 'cc_warning_emails' configuration option is set,
  # add that as a Cc header.
  my $cc ;
  if ($config->cc_warning_emails())
  {
    $cc = $config->cc_warning_emails() ;
  }
  else
  {
    $cc = undef ;
  }


  # Construct the body and send.
  my $body = <<"EOB";
Greetings,

According to our records, the certificate that you purchased for:

    $cn_uppercase

will expire on:

    $expires_at_date_local

When the certificate expires, any client attempting to connect to your SSL
service will issue warnings, and may refuse connection entirely.

If this server is no longer being used, please feel free to ignore this
notification.  If, however, you need to get a new certificate for this
server, please see the instructions at:

<http://www.stanford.edu/services/ssl/>
EOB

  my $subject = "SSL certificate for $cn_uppercase is about to expire" ;

  Stanford::Certificate::Mail::send_email(
    'to'      => $contact_email,
    'cc'      => $cc,
    'from'    => "ITS SSL Certificate Service <$from_email_address>",
    'subject' => $subject,
    'body'    => $body,
  ) ;

  return ;
}

# Look for all LegacyExpiration objects that match $cn and mark those
# found to suppress notifications.
#
# If $cn is set to '*' then loop through
# all the certificates in the Certificates database and for each
# certificate mark any matching legacy expirations to suppress
# notifcations.
#
# Returns the array of LegacyExpiration objects marked as suppress
# notifications.
sub suppress_superseded_legacy_expirations {
    my ($schema, $cn) = @_ ;

    if (!$cn) {
        croak 'missing cn' ;
    }

    if ($cn eq '*') {
        croak 'not yet implemented' ;
    }

    my @legacies =
      Stanford::Certificate::Result::LegacyExpiration::findByCn($schema,
                                                                $cn) ;
    foreach my $legacy (@legacies) {
        my $msg = "suppressing notification for legacy expiration $cn" ;
        Stanford::Certificate::Log::INFO($msg) ;
        $legacy->suppress_notification() ;
        $legacy->update() ;
    }

    return @legacies ;
}


1;

