#<<<
## no critic
use utf8;
package Stanford::Certificate::Result::IncommonStatus;

# Created by DBIx::Class::Schema::Loader
# DO NOT MODIFY THE FIRST PART OF THIS FILE

=head1 NAME

Stanford::Certificate::Result::IncommonStatus

=cut

use strict;
use warnings;

use base 'DBIx::Class::Core';

=head1 TABLE: C<incommon_statuses>

=cut

__PACKAGE__->table("incommon_statuses");

=head1 ACCESSORS

=head2 id

  data_type: 'integer'
  is_auto_increment: 1
  is_nullable: 0

=head2 certificate_id

  data_type: 'integer'
  is_foreign_key: 1
  is_nullable: 1

=head2 status

  data_type: 'enum'
  extra: {list => ["approved","applied","declined","expired","invalid","issued","rejected","requested","revoked","unmanaged"]}
  is_nullable: 0

=head2 status_at

  data_type: 'datetime'
  datetime_undef_if_invalid: 1
  is_nullable: 1

=head2 created_at

  data_type: 'datetime'
  datetime_undef_if_invalid: 1
  is_nullable: 0

=cut

__PACKAGE__->add_columns(
  "id",
  { data_type => "integer", is_auto_increment => 1, is_nullable => 0 },
  "certificate_id",
  { data_type => "integer", is_foreign_key => 1, is_nullable => 1 },
  "status",
  {
    data_type => "enum",
    extra => {
      list => [
        "approved",
        "applied",
        "declined",
        "expired",
        "invalid",
        "issued",
        "rejected",
        "requested",
        "revoked",
        "unmanaged",
      ],
    },
    is_nullable => 0,
  },
  "status_at",
  {
    data_type => "datetime",
    datetime_undef_if_invalid => 1,
    is_nullable => 1,
  },
  "created_at",
  {
    data_type => "datetime",
    datetime_undef_if_invalid => 1,
    is_nullable => 0,
  },
);

=head1 PRIMARY KEY

=over 4

=item * L</id>

=back

=cut

__PACKAGE__->set_primary_key("id");

=head1 UNIQUE CONSTRAINTS

=head2 C<cert_id_status_uq_idx>

=over 4

=item * L</certificate_id>

=item * L</status>

=back

=cut

__PACKAGE__->add_unique_constraint("cert_id_status_uq_idx", ["certificate_id", "status"]);

=head1 RELATIONS

=head2 certificate

Type: belongs_to

Related object: L<Stanford::Certificate::Result::Certificate>

=cut

__PACKAGE__->belongs_to(
  "certificate",
  "Stanford::Certificate::Result::Certificate",
  { id => "certificate_id" },
  {
    is_deferrable => 1,
    join_type     => "LEFT",
    on_delete     => "RESTRICT",
    on_update     => "RESTRICT",
  },
);

## use critic
#>>>

# Created by DBIx::Class::Schema::Loader v0.07042 @ 2016-04-24 12:24:07
# DO NOT MODIFY THIS OR ANYTHING ABOVE! md5sum:hM044ezK5VR5Zqj3edzL4A


# You can replace this text with custom code or comments, and it will be preserved on regeneration

use Stanford::Certificate::Misc qw/ mysql_time_local mysql_time / ;
use Data::Dumper ;

sub new
{
  my ( $class, $attrs ) = @_;

  $attrs->{'created_at'} = mysql_time() unless defined $attrs->{'created_at'};

  my $new = $class->next::method($attrs);
  return $new;
}

# Return a brand new empty Certificate object.
sub new_incommon_status
{
  my ($self) = shift ;
  my $schema = $self->result_source->schema ;
  return $schema->resultset('IncommonStatus')->new({});
}

sub to_s {
  my ($self) = shift ;

  my $id             = $self->id() ;
  my $certificate_id = $self->certificate_id() ;
  my $status         = $self->status() ;
  my $status_at      = $self->status_at() ;
  my $created_at     = $self->created_at() ;

  my $cert = $self->certificate() ;

  my $cn ;
  if ($cert) {
    $cn   = $cert->cn() ;
  }

  if (!$cn) {
      $cn = q{} ;
  }

  if (!$status_at) {
      $status_at = '<NULL>' ;
  }

  if (!defined($id)) {
      $id = '<NONE>' ;
  }

  if (!defined($certificate_id)) {
      $certificate_id = '<NONE>' ;
  }

  return <<"EOS";
id:             $id
certificate id: $certificate_id ($cn)
status:         $status
status time:    $status_at
created_at:     $created_at
EOS
}

sub is_issued {
  my ($self) = shift ;

  return ($self->status() && ($self->status() eq 'issued')) ;
}
sub is_expired {
  my ($self) = shift ;

  return ($self->status() && ($self->status() eq 'expired')) ;
}
sub is_revoked {
  my ($self) = shift ;

  return ($self->status() && ($self->status() eq 'revoked')) ;
}
sub is_declined {
  my ($self) = shift ;

  return ($self->status() && ($self->status() eq 'declined')) ;
}

sub is_approved {
  my ($self) = shift ;

  return ($self->status() && ($self->status() eq 'approved')) ;
}


# Given an array of statuses for a given certificate, return the
# status that is the most current.
#
# Here is the algorithm:
#
#   declined, expired, revoked, issued, approved, requested, applied,
#   rejected, invalid, unmanaged
my %status_to_order = (
  'declined'  =>  1,
  'expired'   =>  2,
  'revoked'   =>  3,
  'rejected'  =>  4,
  'issued'    =>  5,
  'applied'   =>  6,
  'approved'  =>  7,
  'requested' =>  8,
  'invalid'   =>  9,
  'unmanaged' => 10,
    ) ;

# Returns the array sorted by %status_to_order with the lowest
# numbers coming first.
sub sort_statuses {
    my ($statuses_aref) = @_ ;

    my @statuses = @$statuses_aref ;

    my @sorted_statuses = sort {$status_to_order{$a->status()}
                                <=>
                                $status_to_order{$b->status()}}
                          @statuses ;

    return @sorted_statuses ;
}

sub parse_json {
    my $self = shift ;
    my ($json_text) = @_ ;

    use JSON ; # imports encode_json, decode_json, to_json and from_json.
    my $json = JSON->new->allow_nonref ;

    my $cert_href = $json->decode( $json_text );

    if (!exists($cert_href->{'status'})) {
        croak('missing status attribute') ;
    }


    my $status = lc($cert_href->{'status'}) ;

    my @incommon_statuses = () ;

    ##  ## #    ##  ## #    ##  ## #    ##  ## #    ##  ## #    ##  ## #    ##  ## #
    my $parse_date_fref = sub {
        my ($date) = @_ ;
        if ($date =~ m{^(\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d)}) {
            return $1 ;
        } else {
            croak("could not parse date [$date]") ;
        }
    } ;
    ##  ## #    ##  ## #    ##  ## #    ##  ## #    ##  ## #    ##  ## #    ##  ## #

    if ($status eq 'expired') {
        my $expires = $cert_href->{'expires'} ;
        my $expired_at = $parse_date_fref->($expires) ;
        my $incommon_status = $self->new_incommon_status() ;
        $incommon_status->status('expired') ;
        $incommon_status->status_at($expired_at) ;
        push(@incommon_statuses, $incommon_status) ;
    } elsif ($status =~ m{issued|declined|rejected}) {
        my $incommon_status = $self->new_incommon_status() ;
        $incommon_status->status($status) ;
        push(@incommon_statuses, $incommon_status) ;
    } elsif ($status eq 'revoked') {
        my $revoked = $cert_href->{'expires'} ;
        my $revoked_at = $parse_date_fref->($revoked) ;
        my $incommon_status = $self->new_incommon_status() ;
        $incommon_status->status('revoked') ;
        $incommon_status->status_at($revoked_at) ;
        push(@incommon_statuses, $incommon_status) ;
    } else {
        croak("strange status '$status'?!?") ;
    }

    # requested
    my $requested = $cert_href->{'requested'} ;
    my $incommon_status_req = $self->new_incommon_status() ;
    $incommon_status_req->status('requested') ;
    $incommon_status_req->status_at($parse_date_fref->($requested)) ;
    push(@incommon_statuses, $incommon_status_req) ;

    # approved
    if (exists($cert_href->{'approved'})) {
        my $approved  = $cert_href->{'approved'} ;
        my $status_at = $parse_date_fref->($approved) ;
        my $incommon_status_app->status('requested') ;
        $incommon_status_app->status_at($status_at) ;
        push(@incommon_statuses, $incommon_status_app) ;
    }

    # Get the InCommon order id.
    my $order_id ;
    if (!exists($cert_href->{'id'})) {
        croak("missing certificate id") ;
    } else {
        $order_id = $cert_href->{'id'} ;
    }

    return ($order_id, @incommon_statuses) ;

}


1;
