package Stanford::Certificate::StatusReport ;

use strict;
use warnings;
use autodie qw(open close) ;

# Make this a Moose class
use Moose ;
use namespace::autoclean;

use Carp ;
use Stanford::Certificate::Misc qw/
  mysql_time
  mysql_time_local
/ ;

use Stanford::Orange::Util qw/
  trim
/ ;


has report => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

has number_awaiting_approval => (
  is        => 'rw',
  isa       => 'Num',
  required  => 1,
  default   => 0,
) ;

has number_awaiting_collection => (
  is        => 'rw',
  isa       => 'Num',
  required  => 1,
  default   => 0,
) ;

has number_awaiting_ordering => (
  is        => 'rw',
  isa       => 'Num',
  required  => 1,
  default   => 0,
) ;

has using_test => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has schema => (
  is        => 'rw',
  isa       => 'Object',
  required  => 0,
) ;

# Do not show cancelled, declined, or rejected certificate requests.
sub run_report {
    my ($self) = @_ ;

    if (!$self->schema()) {
        croak 'cannot run the report without a schema' ;
    }

    my $schema = $self->schema() ;

    ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###
    my $logger_fun = sub
    {
      my ($msg) = @_ ;
      return Stanford::Certificate::Log::INFO('[status_report] ' . $msg) ;
    } ;
    ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###
    # Return number of results along with formatted lines
    my $format_result_set_fun = sub
    {
      my ($result_set) = @_ ;
      my $number_of_results = 0 ;

      if (scalar($result_set) == 0)
      {
        return ($number_of_results, 'NO REQUESTS FOUND') ;
      }

      my @lines = () ;
      while (my $request = $result_set->next())
      {
        my $id   = $request->id() ;
        my $cn   = $request->cn() ;
        my $date = mysql_time_local($request->created_at()) ;
        my $reissue = $request->reissue() ;
        my $reissue_formatted = ($reissue ? ' (reissue)' : q{} ) ;

        my $line = sprintf('[%s] %5u %s%s', $date, $id, $cn, $reissue_formatted);
        push(@lines, $line) ;
        ++$number_of_results ;
      }

      return ($number_of_results, join("\n", @lines)) ;
    } ;
    ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###  ##   ###

  my $requests_rs ;

  # Section 1. Get all certificate requests that are ordered but approved
  # (that is, have status ordered).
  $requests_rs = $schema->resultset('Request')
                        ->search({
                                   'order_number' => {q{!=}, undef},
                                   'approved'     => 0,
                                   'cancelled'    => 0,
                                   'declined'     => 0,
                                   'rejected'     => 0,
                                 },
                                 {'order_by' => { -asc => 'created_at' }})
  ;

  my ($number_awaiting_approval, $awaiting_approval_section) =
      $format_result_set_fun->($requests_rs) ;

  # Section 2. Get all certificate requests that are new, i.e., awaiting
  # ordering (some might be reissue requests that have not yet been handled.)
  $requests_rs = $schema->resultset('Request')
                        ->search({
                                   'order_number' => undef,
                                   'cancelled'    => 0,
                                   'declined'     => 0,
                                   'rejected'     => 0,
                                 },
                                 {'order_by' => { -asc => 'created_at' }})
  ;

  my ($number_awaiting_ordering, $awaiting_ordering_section) =
      $format_result_set_fun->($requests_rs) ;

  # Section 3. Get all certificate requests that are approved but not collected
  $requests_rs = $schema->resultset('Request')
          ->search({
             'approved'       => 1,
             'certificate_id' => undef,
             'cancelled'      => 0,
             'declined'       => 0,
             'rejected'       => 0,
                 },
             {'order_by' => { -asc => 'created_at' }}
            )
          ;

  my ($number_awaiting_collection, $awaiting_collection_section) =
      $format_result_set_fun->($requests_rs) ;

  my $config = Stanford::Certificate::Config::instance() ;

  my $today = mysql_time_local(mysql_time()) ;

  my @connect_info = @{ $schema->connection()->storage()->_connect_info() } ;
  my $dsn = $connect_info[0] ;

  my $REPORT = <<"EOR";
SSL Certificate Request Daily Report [generated $today]

Requests awaiting APPROVAL
date ordered          db id subject name
--------------------- ----- ---------------------------
$awaiting_approval_section

Requests awaiting ORDERING
date ordered          db id subject name
--------------------- ----- ---------------------------
$awaiting_ordering_section

Requests awaiting COLLECTION
date ordered          db id subject name
--------------------- ----- ---------------------------
$awaiting_collection_section

(data source name: $dsn)
EOR

  $self->report(trim($REPORT)) ;

  $self->number_awaiting_approval($number_awaiting_approval) ;
  $self->number_awaiting_collection($number_awaiting_collection) ;
  $self->number_awaiting_ordering($number_awaiting_ordering) ;

  return ;
}

__PACKAGE__->meta->make_immutable;

1;
