package Stanford::Certificate::X509 ;

use strict;
use warnings;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
  parse_x509_pem
);

use Carp ;
use Crypt::X509 ;
use MIME::Base64;

use Stanford::Orange::Util qw/ trim tempfile slurp spew run_command / ;
use Stanford::Certificate::Misc qw / mysql_time / ;

sub decode_rsa_pubkey
{
  my ($pubkey_asn) = @_ ;

  my $asn = Convert::ASN1->new() ;
  $asn->prepare(<<'ASN1');

RSAPublicKey ::= SEQUENCE {
      modulus            INTEGER, -- n
      publicExponent     INTEGER  -- e --
      }
ASN1

  $asn->find('RSAPublicKey') ;
  my $result = $asn->decode($pubkey_asn) ;
  return $result ;
}

sub decode_dsa_pubkey
{
  my ($pubkey_asn) = @_ ;

  my $asn = Convert::ASN1->new() ;
  $asn->prepare(<<'ASN1');

DSAPublicKey ::= INTEGER -- public key, Y

ASN1

  $asn->find('DSAPublicKey') ;
  my $result = $asn->decode($pubkey_asn) ;
  return $result ;
}

sub pem_to_der
{
  my ($pem_string) = @_ ;

  # 1. Strip off the headers
  $pem_string =~ s{-----BEGIN\ CERTIFICATE-----}{}xms ;
  $pem_string =~ s{-----END\ CERTIFICATE-----}{}xms ;
  $pem_string = trim($pem_string) ;

  # 2. Convert from BASE64
  my $der_string = decode_base64($pem_string) ;

  return $der_string ;
}

# Given a command name (e.g., "myserver.stanford.edu"), and paths to where
# the CSR and key should be placed, creates a CSR and key.
# Returns the CSR as a string.
sub make_csr
{
  my ($cn, $filename_csr, $filename_key) = @_ ;

  # openssl req -new -nodes -newkey rsa:2048 -keyout hostname.key -out hostname.csr
  # Step 1. Create a "conf" file
  my $conf_file_string = <<"EOC";
[req]
distinguished_name = req_distinguished_name

[req_distinguished_name]
countryName                     = Country Name (2 letter code)
stateOrProvinceName             = State or Province Name (full name)
localityName                    = Locality Name (city, district)
organizationName                = Organization Name (company)
organizationalUnitName          = Organizational Unit Name (department, division)
commonName                      = Common Name (hostname, IP, or your name)

countryName_default             = US
stateOrProvinceName_default     = California
localityName_default            = Stanford
organizationName_default        = Stanford University
organizationalUnitName_default  = IT Services
commonName_default              = $cn
EOC

  my $conf_file = tempfile() ;
  spew($conf_file_string, $conf_file) ;
  my $cmd = "openssl req -new -nodes -newkey rsa:2048 -keyout $filename_key" .
            " -out $filename_csr -batch -config $conf_file" ;

  my ($stdout, $stderr) = Capture::Tiny::capture {
     `$cmd`
  };

  unlink($conf_file)
    or croak "could not unlink file '$conf_file'" ;

  my $csr = slurp($filename_csr) ;
  return $csr ;
}

sub parse_x509_pem_file
{
  my ($filename) = @_ ;
  my $pem_string = slurp($filename) ;
  return parse_x509_pem($pem_string) ;
}

sub x509_date_string_to_MySQL
{
  my ($x509_date_string_raw) = @_ ;

  # May 19 23:59:59 2014 GMT
  if ($x509_date_string_raw =~ m{^(\D+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\d\d\d\d)\s+GMT$}x)
  {
    my $time_t = str2time($x509_date_string_raw);
    return mysql_time($time_t) ;
  }
  else
  {
    croak "[x509_date_string_to_MySQL] could not parse date '$x509_date_string_raw'" ;
  }
}

# Given a string containing zero, one, or many PEM strings, return
# an array containing the PEM strings, one in each array position.
#
# If the string contains anything other than complete PEM string(s) (or
# white-space), this function will croak.
#
# Returns the empty array if the string is empty (or all white-space).
sub split_pems_string
{
  my ($pems_string) = @_ ;

  my @pems = () ;

  $pems_string = trim($pems_string) ;
  while ($pems_string)
  {
    if ($pems_string =~ s{(-----BEGIN\ CERTIFICATE-----.*?-----END\ CERTIFICATE-----)}{}xms)
    {
      my $current_pem = $1 ;

      push(@pems, $current_pem) ;
      $pems_string = trim($pems_string) ;
    }
    else
    {
      croak "Found something that is not a certificate: [$pems_string]\n" ;
    }
  }

  return @pems ;
}


# Returns a hash with these keys:
#                                   Aliases
#    issuer                         issuer_cn
#    subject
#    cn                             subject_cn, CN
#    not_before                     valid_at
#    not_after                      expires_at
#    serial_number                  serial
#    key_size                       key_length
#    pubkey_algorithm               key_algorithm
#    version
#    subject_key_identifier
#    authority_key_identifier
#    subject_alternative_names      sans
#    ca                             CA
#    signature_algorithm
#
# The subject_alternative_names value is a hash reference. Each
# key of the hash is a subject alternative name (e.g., 'DNS', 'IP',
# 'email', etc.). The value is an array reference of subject alternative
# names of that type. Example:
#
# {
#   'DNS' => [ 'mydomain.com', 'mydomain.net', ],
#   'IP'  => [ '100.101.102.103', ],
# }
#
#  NOTE 1. not_before and not_after (and their aliases) are
#          MySQL-compatible time strings in GMT (UTC) time.
sub parse_x509_pem
{
  my ($x509_pem_string) = @_ ;

  if (!$x509_pem_string)
  {
    croak "null or empty string passed to parse_x509_pem()" ;
  }

  my %field_to_value = () ;

  my $der_string = pem_to_der($x509_pem_string) ;

  if (!$der_string)
  {
    croak "no DER string derived from PEM string" ;
  }

  my $decoded = Crypt::X509->new( cert => $der_string );

  if ($decoded->error)
  {
    croak "Error on parsing PEM string:" . $decoded->error ;
  }

  # 1. Get version.
  if (defined($decoded->version()))
  {
    $field_to_value{'version'} = _get_version($decoded->version_string()) ;
  }
  else
  {
    # Default is version 1
    $field_to_value{'version'} = 1 ;
  }

  # 2. Get Serial Number (normalize it by coverting to upper-case hex).
  # Example: 7F71C1D3A226B0D2B113F3E68167643E
  if (defined($decoded->serial()))
  {
    my $serial_bigint = $decoded->serial() ;
    my $hex_string ;
    if (ref($serial_bigint) eq 'Math::BigInt')
    {
      $hex_string = $serial_bigint->as_hex() ;
    }
    else
    {
      $hex_string = sprintf("%x", $serial_bigint) ;
    }
    $hex_string =~ s{^0x}{}x ;
    $hex_string = uc($hex_string) ;
    $field_to_value{'serial_number'} = $hex_string;
  }

  # 3. Issuer DN
  if (defined($decoded->Issuer))
  {
    my @issuers = @{ $decoded->Issuer } ;
    my $issuers = join(", ", @issuers) ;

    $field_to_value{'issuer'} = $issuers;
  }

  # 4. Subject DN
  if (defined($decoded->Subject()))
  {
    my @subjects = @{ $decoded->Subject } ;
    my @subjects_formatted = () ;

    # Clean-up
    foreach my $subject (@subjects)
    {
      $subject =~ s{^(2\.5\.4\.5)(=.*)$}{serialNumber$2}x ;
      $subject =~ s{^(2\.5\.4\.9)(=.*)$}{streetAddress$2}x ;
      $subject =~ s{^(2\.5\.4\.17)(=\d+)$}{postalCode$2}x ;
      $subject =~ s{^l(=.*)$}{L$1}x ;
      $subject =~ s{^s(=.*)$}{ST$1}ix ;
      push(@subjects_formatted, $subject) ;
    }
    # Fix up a couple of the OIDs.
    my $subjects = join(", ", @subjects_formatted) ;

    $field_to_value{'subject'} = $subjects ;
  }

  # 5. Common Name
  if (defined($decoded->subject_cn()))
  {
    $field_to_value{'cn'} = $decoded->subject_cn() ;
  }

  # 6. Subject and Authority Key Identifier
  if (defined($decoded->subject_keyidentifier()))
  {
    $field_to_value{'subject_key_identifier'} =
      uc(unpack('H*', $decoded->subject_keyidentifier())) ;
  }

  if (defined($decoded->key_identifier()))
  {
    $field_to_value{'authority_key_identifier'} =
      uc(unpack('H*', $decoded->key_identifier())) ;
  }

  # 7. Not Before and Not After
  if (defined($decoded->not_before()))
  {
    $field_to_value{'not_before'} = mysql_time($decoded->not_before()) ;
  }

  if (defined($decoded->not_after()))
  {
    $field_to_value{'not_after'} = mysql_time($decoded->not_after()) ;
  }

  # 8. Public-key algorithm.
  #    Will be one of these:
  #        * RSA
  #        * DSA
  #        * UNKNOWN
  if (defined($decoded->pubkey_algorithm()))
  {
    my $pubkey_algorithm_oid = $decoded->pubkey_algorithm() ;
    $field_to_value{'pubkey_algorithm_oid'} = $pubkey_algorithm_oid ;

    if ($pubkey_algorithm_oid eq '1.2.840.113549.1.1.1')
    {
      $field_to_value{'pubkey_algorithm'} = 'RSA' ;
    }
    elsif ($pubkey_algorithm_oid eq '1.2.840.10040.4.1')
    {
      $field_to_value{'pubkey_algorithm'} = 'DSA' ;
    }
    else
    {
      $field_to_value{'pubkey_algorithm'} = 'UNKNOWN' ;
    }
  }

  # 9. Key size
  if (defined($decoded->pubkey()))
  {
    if ($field_to_value{'pubkey_algorithm'} eq 'RSA')
    {
      my $pubkey_and_exponent_asn = $decoded->pubkey() ;
      my $pubkey_and_exponent_decoded = decode_rsa_pubkey($pubkey_and_exponent_asn) ;
      my $modulus = uc($pubkey_and_exponent_decoded->{'modulus'}->as_hex()) ;
      $modulus =~ s{^0x}{}ix ;
      $field_to_value{'modulus'} = $modulus ;
      $field_to_value{'key_size'} = 4 * length($modulus) ;
    }
    elsif ($field_to_value{'pubkey_algorithm'} eq 'DSA')
    {
      my $pubkey_and_exponent_asn = $decoded->pubkey() ;
      my $pubkey_and_exponent_decoded = decode_dsa_pubkey($pubkey_and_exponent_asn) ;
      my $public_key = uc($pubkey_and_exponent_decoded->as_hex()) ;
      $public_key =~ s{^0x}{}ix ;
      $field_to_value{'key_size'} = 4 * length($public_key) ;
    }
  }

  # 10. Subject Alternative Names
  #   otherName                       [0]     AnotherName,
  #   rfc822Name                      [1]     IA5String,
  #   dNSName                         [2]     IA5String,
  #   x400Address                     [3]     ORAddress,
  #   directoryName                   [4]     Name,
  #   ediPartyName                    [5]     EDIPartyName,
  #   uniformResourceIdentifier       [6]     IA5String,
  #   iPAddress                       [7]     OCTET STRING,
  #   registeredID
  #
  if (defined($decoded->SubjectAltName()))
  {
    my @sans = @{ $decoded->SubjectAltName() } ;

    my $san_type_to_values_href = {} ;
    foreach my $san (@sans)
    {
      if ($san =~ m{^([^=]+?)=(.*)$}x)
      {
        my $san_type = trim($1) ;
        my $san_name = trim($2) ;
        if (!exists($san_type_to_values_href->{$san_type}))
        {
          $san_type_to_values_href->{$san_type} = [] ;
        }
        push(@{$san_type_to_values_href->{$san_type}}, $san_name) ;
      }
    }

    $field_to_value{'subject_alternative_names'} = $san_type_to_values_href ;
  }

  # 11. CA (certificate authority cert)
  $field_to_value{'ca'} = 0 ;
  if (defined($decoded->BasicConstraints()))
  {
    my @BasicConstraints = @{ $decoded->BasicConstraints() } ;
    # Look for 'cA = 1'
    foreach my $basic_constraint (@BasicConstraints)
    {
      if ($basic_constraint =~ m{cA\ =\ 1}x)
      {
        $field_to_value{'ca'} = 1 ;
      }
    }
  }

  # 12. Key Usage. Convert to a hash. Possible values:
  #   digitalSignature
  #   nonRepudiation
  #   keyEncipherment
  #   dataEncipherment
  #   keyAgreement
  #   keyCertSign  (this one means the cert can sign other certs)
  #   cRLSign
  my $key_usage_href = {} ;
  my $keyUsage = $decoded->KeyUsage() ;
  if ($keyUsage)
  {
    my @keyUsages = @$keyUsage ;
    foreach my $key_usage (@keyUsages)
    {
      $key_usage_href->{$key_usage} = 1 ;
    }
  }
  $field_to_value{'key_usage_href'} = $key_usage_href ;

  # 13. Signature algorithm (see RFC 3279 and
  # http://msdn.microsoft.com/en-us/library/ff635603.aspx.)
  $field_to_value{'signature_algorithm'} = undef ;
  if (defined($decoded->sig_algorithm())) {
      my $sig_algorithm_oid = $decoded->sig_algorithm() ;

      my %oid_to_string = (
          '1.2.840.113549.1.1.2'  => 'md2WithRSAEncryption',
          '1.2.840.113549.1.1.4'  => 'md5WithRSAEncryption',
          '1.2.840.113549.1.1.5'  => 'sha1WithRSAEncryption',
          '1.2.840.113549.1.1.11' => 'sha256WithRSAEncryption',
          '1.2.840.113549.1.1.12' => 'sha384WithRSAEncryption',
          '1.2.840.113549.1.1.13' => 'sha512WithRSAEncryption',
          '1.2.840.10045.4.3.3'   => 'ecdsa-with-SHA384',

      ) ;

      if (exists($oid_to_string{$sig_algorithm_oid})) {
          my $sig_algorithm_string = $oid_to_string{$sig_algorithm_oid} ;
          $field_to_value{'signature_algorithm'} = $sig_algorithm_string ;
      }
  }

  # 14. OCSP URI (only if openssl is installed).
  my $ocsp_uri ;
  eval
  {
    my $der_temp_file = tempfile() ;
    spew($der_string, $der_temp_file) ;
    my $cmd = "openssl x509 -in $der_temp_file -inform DER -noout " .
              "-ocsp_uri" ;

    my ($stdout, $stderr, $exit_value) = run_command($cmd) ;

    unlink $der_temp_file
      or croak "could not unlink temp file $der_temp_file: $!" ;

    if (trim($stderr))
    {
      croak "error determining OCSP URI: " . trim($stderr) ;
    }

    $ocsp_uri = trim($stdout) ;
  } ;

  if (!$@)
  {
    if (!defined($ocsp_uri) || ($ocsp_uri eq q{}))
    {
      $ocsp_uri = undef ;
    }
    $field_to_value{'ocsp_uri'} = $ocsp_uri ;
  }
  else
  {
    $field_to_value{'ocsp_uri'} = undef ;
  }


  # ALIASES
  $field_to_value{'subject_cn'}    = $field_to_value{'cn'} ;
  $field_to_value{'CN'}            = $field_to_value{'cn'} ;
  $field_to_value{'valid_at'}      = $field_to_value{'not_before'} ;
  $field_to_value{'expires_at'}    = $field_to_value{'not_after'} ;
  $field_to_value{'serial'}        = $field_to_value{'serial_number'} ;
  $field_to_value{'sans'}          = $field_to_value{'subject_alternative_names'} ;
  $field_to_value{'key_length'}    = $field_to_value{'key_size'} ;
  $field_to_value{'key_algorithm'} = $field_to_value{'pubkey_algorithm'} ;
  $field_to_value{'CA'}            = $field_to_value{'ca'} ;

  return %field_to_value ;
}

sub _get_version
{
  my ($decoded_string) = @_ ;

  my %version_string_to_int = (
      'v1' => 1,
      'v2' => 2,
      'v3' => 3,
    ) ;
  return $version_string_to_int{$decoded_string} ;
}


# Given an array (reference) of certificate PEM strings, find the one that
# is not a signing certificate. Return an array with the first element
# being the PEM string that is not a signing certificate and with the rest
# of the PEM strings filling out the rest of the array.
#
# Example: assume that we are pased [ $p1, $p2, $p3] where $p2 is not a
# signing certificate. This function will return [$p2, $p1, $p3].
#
# If there are _more_ than one certificates that are not signing
# certificates, this function will raise an exception.
#
# If none of the certificates are not signing certificates, the first
# element of the array will be undef.
#
# Example. Assume that
#  $p1 is a PEM string representing the certificate for the server myserver.stanford,
#  $p2 is a PEM string representing the certificate that signed $p1, and
#  $p3 is a PEM string presenting a root certificate
#
#  Input             Output
#  [$p1, $p2, $p3]   [$p1, $p2, $p3]
#  [$p2, $p1, $p3]   [$p1, $p2, $p3]
#  [$p1, $p1, $p3]   exception (more than one non-signing cert)
#  [$p2, $p2, $p3]   [undef, $p2, $p2, $p3]
#  []                []
sub separate_pems
{
  my ($pem_strings_aref) = @_ ;

  my @pem_strings = @$pem_strings_aref ;

  if (scalar(@pem_strings) == 0)
  {
    return () ;
  }

  my @signing_certs = () ;
  my $server_cert = undef ;
  foreach my $pem_string (@pem_strings)
  {
    my %field_to_value = parse_x509_pem($pem_string) ;
    my $key_usage_href = $field_to_value{'key_usage_href'} ;
    my %key_usage = %$key_usage_href ;

    if (exists($key_usage{'keyCertSign'}))
    {
      # This is a signing cert
      push (@signing_certs, $pem_string) ;
    }
    else
    {
      if (defined($server_cert))
      {
        # $server_cert already defined!! Bad!!
        croak "found more than one server cert" ;
      }
      else
      {
        $server_cert = $pem_string ;
      }
    }
  }

  # Combine the server and signing certs (make sure to put the server cert
  # at the front).
  unshift(@signing_certs, $server_cert) ;

  return @signing_certs ;
}

# Given an array (reference) of PEM strings, save in the certificates
# database.
sub save_pems
{
  my ($pem_strings_aref, $schema) = @_ ;

  # Add each certificate to the certificates table. Note that if the
  # certificate already exists in the table, we don't add it. Recall that
  # every X509 certificate is (supposed to be) uniquely determined by it's
  # issuer and serial number.
  my $cert = $schema->resultset('Certificate')->new({});
  my @certs = $cert->make_certs_from_pem_string($pem_strings_aref) ;
  return @certs ;
}



1;
